struct gifhead {char ID[6];short int wide,high;char gidb,backcol,unbek;};
struct imghead {short int dx,dy,br,hi;char lidb;};
typedef struct paltable { unsigned char r,g,b; } dregs[256];

int GIF_saveghead(int handle,int br,int h,int bpp,dregs *pt);
int GIF_savelhead(int handle,int br,int h,int bpp,dregs *pt);
int GIF_saveimg  (int handle,int bpp,int getpixel(void));
int GIF_saveend  (int handle);
int GIF_save(char *name,int br,int h,dregs *pt,int bpp,int getpixel(void));
